#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <float.h>
#include "common/xc.h"
#include "libxc/gga_x_pbe.h"
#include "genxc/gga_x_pbe_genxc.h"

// Precision for correctness checking
#define EPSILON     0.00000001

int nearlyEqual(double a, double b) {
    //double abs_a = abs(a);
    //double abs_b = abs(b);
    double diff = abs(a - b);

    //if(a == b) { return 1; }
    //if(a == 0 || b == 0 || abs_a + abs_b < MIN_NORMAL) { return diff < (EPSILON * MIN_NORMAL); }
    //return (diff / min((abs_a + abs_b), DBL_MAX)) < EPSILON;
    return diff < EPSILON;
}

double getTimeStamp() {
    struct timespec ts;
    clock_gettime(CLOCK_MONOTONIC, &ts);
    return (double)ts.tv_sec + (double)ts.tv_nsec * 1.e-9;
}

int main() {
  xc_func_type gga_libxc, gga_genxc;
  int i;
  int npoints = 10*1000;
  //int npoints = 100;
  double *rho_libxc, *sigma_libxc;
  double *rho_genxc, *sigma_genxc;
  double *zk_libxc, *vrho_libxc, *vsigma_libxc;
  double *zk_genxc, *vrho_genxc, *vsigma_genxc;
  double *v2rho2_libxc, *v2rhosigma_libxc, *v2sigma2_libxc;
  double *v2rho2_genxc, *v2rhosigma_genxc, *v2sigma2_genxc;
  double *v3rho3_libxc, *v3rho2sigma_libxc, *v3rhosigma2_libxc, *v3sigma3_libxc;
  double *v3rho3_genxc, *v3rho2sigma_genxc, *v3rhosigma2_genxc, *v3sigma3_genxc;
  double *v4rho4_libxc, *v4rho3sigma_libxc, *v4rho2sigma2_libxc, *v4rhosigma3_libxc, *v4sigma4_libxc;
  double *v4rho4_genxc, *v4rho3sigma_genxc, *v4rho2sigma2_genxc, *v4rhosigma3_genxc, *v4sigma4_genxc;
  double S, E;

  rho_libxc          = (double*) malloc( 2*npoints*sizeof(double));
  sigma_libxc        = (double*) malloc( 3*npoints*sizeof(double));
  zk_libxc           = (double*) malloc( 1*npoints*sizeof(double));
  vrho_libxc         = (double*) malloc( 2*npoints*sizeof(double));
  vsigma_libxc       = (double*) malloc( 3*npoints*sizeof(double));
  v2rho2_libxc       = (double*) malloc( 3*npoints*sizeof(double));
  v2rhosigma_libxc   = (double*) malloc( 6*npoints*sizeof(double));
  v2sigma2_libxc     = (double*) malloc( 6*npoints*sizeof(double));
  rho_genxc          = (double*) malloc( 2*npoints*sizeof(double));
  sigma_genxc        = (double*) malloc( 3*npoints*sizeof(double));
  zk_genxc           = (double*) malloc( 1*npoints*sizeof(double));
  vrho_genxc         = (double*) malloc( 2*npoints*sizeof(double));
  vsigma_genxc       = (double*) malloc( 3*npoints*sizeof(double));
  v2rho2_genxc       = (double*) malloc( 3*npoints*sizeof(double));
  v2rhosigma_genxc   = (double*) malloc( 6*npoints*sizeof(double));
  v2sigma2_genxc     = (double*) malloc( 6*npoints*sizeof(double));
  #ifdef KXC
  v3rho3_libxc       = (double*) malloc( 4*npoints*sizeof(double));
  v3rho2sigma_libxc  = (double*) malloc( 9*npoints*sizeof(double));
  v3rhosigma2_libxc  = (double*) malloc(12*npoints*sizeof(double));
  v3sigma3_libxc     = (double*) malloc(10*npoints*sizeof(double));
  v3rho3_genxc       = (double*) malloc( 4*npoints*sizeof(double));
  v3rho2sigma_genxc  = (double*) malloc( 9*npoints*sizeof(double));
  v3rhosigma2_genxc  = (double*) malloc(12*npoints*sizeof(double));
  v3sigma3_genxc     = (double*) malloc(10*npoints*sizeof(double));
  #else
  v3rho3_libxc       = (double*) NULL;
  v3rho2sigma_libxc  = (double*) NULL;
  v3rhosigma2_libxc  = (double*) NULL;
  v3sigma3_libxc     = (double*) NULL;
  v3rho3_genxc       = (double*) NULL;
  v3rho2sigma_genxc  = (double*) NULL;
  v3rhosigma2_genxc  = (double*) NULL;
  v3sigma3_genxc     = (double*) NULL;
  #endif
  #ifdef LXC
  v4rho4_libxc       = (double*) malloc( 5*npoints*sizeof(double));
  v4rho3sigma_libxc  = (double*) malloc(12*npoints*sizeof(double));
  v4rho2sigma2_libxc = (double*) malloc(18*npoints*sizeof(double));
  v4rhosigma3_libxc  = (double*) malloc(20*npoints*sizeof(double));
  v4sigma4_libxc     = (double*) malloc(15*npoints*sizeof(double));
  v4rho4_genxc       = (double*) malloc( 5*npoints*sizeof(double));
  v4rho3sigma_genxc  = (double*) malloc(12*npoints*sizeof(double));
  v4rho2sigma2_genxc = (double*) malloc(18*npoints*sizeof(double));
  v4rhosigma3_genxc  = (double*) malloc(20*npoints*sizeof(double));
  v4sigma4_genxc     = (double*) malloc(15*npoints*sizeof(double));
  #else
  v4rho4_libxc       = (double*) NULL;
  v4rho3sigma_libxc  = (double*) NULL;
  v4rho2sigma2_libxc = (double*) NULL;
  v4rhosigma3_libxc  = (double*) NULL;
  v4sigma4_libxc     = (double*) NULL;
  v4rho4_genxc       = (double*) NULL;
  v4rho3sigma_genxc  = (double*) NULL;
  v4rho2sigma2_genxc = (double*) NULL;
  v4rhosigma3_genxc  = (double*) NULL;
  v4sigma4_genxc     = (double*) NULL;
  #endif

  if(xc_func_init(&gga_genxc, XC_GGA_X_BCGP_GENXC, XC_POLARIZED) < 0) {
    fprintf(stderr, "XC_GGA_X_BCGP_GENXC not found!\n");
    return -1;
  }

  for(i=0; i<npoints; i++){
    rho_genxc[2*i + 0]   = 0.048 + i/(double)(npoints);
    rho_genxc[2*i + 1]   = 0.025;
    sigma_genxc[3*i + 0] = 0.0046;
    sigma_genxc[3*i + 1] = 0.0044;
    sigma_genxc[3*i + 2] = 0.0041;
  }

  S = getTimeStamp();
  xc_gga(
    &gga_genxc, npoints, rho_genxc, sigma_genxc, zk_genxc, vrho_genxc, vsigma_genxc, v2rho2_genxc, v2rhosigma_genxc, v2sigma2_genxc,
    v3rho3_genxc, v3rho2sigma_genxc, v3rhosigma2_genxc, v3sigma3_genxc, v4rho4_genxc, v4rho3sigma_genxc, v4rho2sigma2_genxc,
    v4rhosigma3_genxc, v4sigma4_genxc
  );
  E = getTimeStamp();
  fprintf(stderr, "GGA_X_PBE_GENXC: %.4fms\n", 1e3*(E-S));

  if(xc_func_init(&gga_libxc, XC_GGA_X_BCGP, XC_POLARIZED) < 0) {
    fprintf(stderr, "XC_GGA_X_BCGP not found!\n");
    return -1;
  }

  for(i=0; i<npoints; i++){
    rho_libxc[2*i + 0]   = 0.048 + i/(double)(npoints);
    rho_libxc[2*i + 1]   = 0.025;
    sigma_libxc[3*i + 0] = 0.0046;
    sigma_libxc[3*i + 1] = 0.0044;
    sigma_libxc[3*i + 2] = 0.0041;
  }

  S = getTimeStamp();
  xc_gga(
    &gga_libxc, npoints, rho_libxc, sigma_libxc, zk_libxc, vrho_libxc, vsigma_libxc, v2rho2_libxc, v2rhosigma_libxc, v2sigma2_libxc,
    v3rho3_libxc, v3rho2sigma_libxc, v3rhosigma2_libxc, v3sigma3_libxc, v4rho4_libxc, v4rho3sigma_libxc, v4rho2sigma2_libxc,
    v4rhosigma3_libxc, v4sigma4_libxc
  );
  E = getTimeStamp();
  fprintf(stderr, "GGA_X_PBE: %.4fms\n", 1e3*(E-S));

  #define CHECK_CORRECTNESS(arr1,arr2,i);  if(!nearlyEqual(arr1[i], arr2[i])) { fprintf(stderr, "Results differ!\n"); return -1; }
  for(i = 0; i < npoints; i++) {
      CHECK_CORRECTNESS(zk_libxc, zk_genxc, i);
//    CHECK_CORRECTNESS(rho_libxc, rho_genxc, 2 * i + 0);
      CHECK_CORRECTNESS(vrho_libxc, vrho_genxc, 2 * i + 0);
      CHECK_CORRECTNESS(vrho_libxc, vrho_genxc, 2 * i + 1);
//    CHECK_CORRECTNESS(v2rho2_libxc, v2rho2_genxc, 3 * i + 0);
      CHECK_CORRECTNESS(vsigma_libxc, vsigma_genxc, 3 * i + 0);
      CHECK_CORRECTNESS(vsigma_libxc, vsigma_genxc, 3 * i + 1);
      CHECK_CORRECTNESS(vsigma_libxc, vsigma_genxc, 3 * i + 2);
  }
    printf("Results match\n");

  xc_func_end(&gga_libxc);
  xc_func_end(&gga_genxc);
  return 0;
}
