import sympy as sp

# import math functions and constants directly to allow switching to implementations of other packages easily
from sympy import log, sqrt

# rs, z, t = sp.symbols('rs, z, t')
# rs, z, xt, xs0, xs1 = sp.symbols('rs, z, xt, xs0, xs1')

p_a_zeta_threshold = sp.Symbol('p_a_zeta_threshold')
p_a_dens_threshold = sp.Symbol('p_a_dens_threshold')


# p->zeta_threshold

# define my_piecewise3(c, x1, x2) ((c) ? (x1) : (x2))
def my_piecewise3(cond, left, right):
    # print(cond)
    pw = sp.Piecewise((left, cond), (right, True))
    # print(pw)
    # print(sp.simplify(pw))
    # print()
    # return sp.simplify(pw)
    return pw


# define my_piecewise5(c1, x1, c2, x2, x3) ((c1) ? (x1) : ((c2) ? (x2) : (x3)))
def my_piecewise5(c1, x1, c2, x2, x3):
    return sp.Piecewise((x1, c1), (x2, c2), (x3, True))


# from https://gitlab.com/libxc/libxc/-/blob/master/maple/util.mpl

X2S = 1 / (2 * (6 * sp.pi ** 2) ** sp.S(1 / 3))
X2S_2D = 1 / (2 * (4 * sp.pi) ** sp.S(1 / 2))

MU_GE = sp.S(10 / 81)
X_FACTOR_C = 3 / 8 * (3 / sp.pi) ** (1 / 3) * 4 ** (2 / 3)
X_FACTOR_2D_C = 8 / (3 * sp.sqrt(sp.pi))
K_FACTOR_C = 3 / 10 * (6 * sp.pi ** 2) ** (2 / 3)

# # 1D
# DIMENSIONS = 1
# RS_FACTOR = 1 / 2
# # 2D
# DIMENSIONS = 2
# RS_FACTOR = 1 / sqrt(sp.pi)
# LDA_X_FACTOR = -X_FACTOR_2D_C
# 3D
DIMENSIONS = 3
RS_FACTOR = sp.S(3 / (4 * sp.pi)) ** sp.S(1 / 3)
LDA_X_FACTOR = -X_FACTOR_C


def opz_pow_n(z, n):
    return my_piecewise3(1 + z <= p_a_zeta_threshold, p_a_zeta_threshold ** n, (1 + z) ** n)


def f_zeta(z):
    return (opz_pow_n(z, sp.S(4 / 3)) + opz_pow_n(-z, sp.S(4 / 3)) - 2) / (2 ** sp.S(4 / 3) - 2)


def mphi(z):
    return (opz_pow_n(z, sp.S(2 / 3)) + opz_pow_n(-z, sp.S(2 / 3))) / 2


def tt(rs, z, xt):
    return xt / (4 * 2 ** sp.S(1 / 3) * mphi(z) * sqrt(rs))


def r_ws(n):
    return RS_FACTOR / n ** sp.S(1 / DIMENSIONS)


def n_total(rs):
    return (RS_FACTOR / rs) ** DIMENSIONS


def n_spin(rs, z):
    return (1 + z) * n_total(rs) / 2


# Screening for extreme values of zeta
def z_thr(z):
    return my_piecewise5(
        1 + z <= p_a_zeta_threshold, p_a_zeta_threshold - 1,
        1 - z <= p_a_zeta_threshold, 1 - p_a_zeta_threshold,
        z)


# Generate exchange and kinetic functionals from the expression for the
# enhancement factor
def lda_x_spin(rs, z):
    return LDA_X_FACTOR * opz_pow_n(z, 1 + 1 / DIMENSIONS) * 2 ** (-1 - 1 / DIMENSIONS) * (RS_FACTOR / rs)


def lda_k_spin(rs, z):
    return K_FACTOR_C * opz_pow_n(z, 5 / 3) * 2 ** (-5 / 3) * (RS_FACTOR / rs) ** 2


# Screen out small densities as well as the zero component from fully spin polarized densities due to terms of the form (1+z)^{-n}
def screen_dens(rs, z):
    return n_spin(rs, z) <= p_a_dens_threshold


def gga_exchange(func, rs, z, xs0, xs1):
    return (my_piecewise3(screen_dens(rs, z), 0, lda_x_spin(rs, z_thr(z)) * func(xs0))
            + my_piecewise3(screen_dens(rs, -z), 0, lda_x_spin(rs, z_thr(-z)) * func(xs1)))


# from https://gitlab.com/libxc/libxc/-/blob/master/maple/lda_exc/lda_c_pw.mpl

# add None to honor Maple numbering starting with 1
params_a_pp = [None, 1, 1, 1]
params_a_a = [None, sp.S(0.031091), sp.S(0.015545), sp.S(0.016887)]
params_a_alpha1 = [None, sp.S(0.21370), sp.S(0.20548), sp.S(0.11125)]
params_a_beta1 = [None, sp.S(7.5957), sp.S(14.1189), sp.S(10.357)]
params_a_beta2 = [None, sp.S(3.5876), sp.S(6.1977), sp.S(3.6231)]
params_a_beta3 = [None, sp.S(1.6382), sp.S(3.3662), sp.S(0.88026)]
params_a_beta4 = [None, sp.S(0.49294), sp.S(0.62517), sp.S(0.49671)]
params_a_fz20 = sp.S(1.709921)

params_a_a = [None, sp.S(0.0310907), sp.S(0.01554535), sp.S(0.0168869)]
params_a_fz20 = sp.S(1.709920934161365617563962776245)


# (* Equation (10) *)
def g_aux(k, rs):
    return params_a_beta1[k] * sqrt(rs) + params_a_beta2[k] * rs + params_a_beta3[k] * rs ** sp.S(1.5) + params_a_beta4[k] * rs ** (params_a_pp[k] + 1)


def g(k, rs):
    return -2 * params_a_a[k] * (1 + params_a_alpha1[k] * rs) * log(1 + 1 / (2 * params_a_a[k] * g_aux(k, rs)))


# (* Equation (8) *)
# (* Attention, the function g parametrizes -alpha *)
def f_pw(rs, zeta):
    return g(1, rs) + zeta ** 4 * f_zeta(zeta) * (g(2, rs) - g(1, rs) + g(3, rs) / params_a_fz20) - f_zeta(zeta) * g(3, rs) / params_a_fz20


# from https://gitlab.com/libxc/libxc/-/blob/master/maple/gga_exc/gga_c_pbe.mpl

# def mbeta(rs, t):
#     return 0.06672455060314922
#
#
# mgamma = (1 - log(2)) / pi ** 2
# BB = 1
#
#
# def tp(rs, z, xt):
#     return tt(rs, z, xt)
#
#
# # (* Equation (8) *)
# def A(rs, z, t):
#     return mbeta(rs, t) / (mgamma * (exp(-f_pw(rs, z) / (mgamma * mphi(z) ** 3)) - 1))
#
#
# # (* Equation (7) *) helpers
# def f1(rs, z, t):
#     return t ** 2 + BB * A(rs, z, t) * t ** 4
#
#
# def f2(rs, z, t):
#     return mbeta(rs, t) * f1(rs, z, t) / (mgamma * (1 + A(rs, z, t) * f1(rs, z, t)))
#
#
# # (* Equation (7) *)
# def fH(rs, z, t):
#     return mgamma * mphi(z) ** 3 * log(1 + f2(rs, z, t))
#
#
# def f_pbe(rs, z, xt, xs0, xs1):
#     return f_pw(rs, z) + fH(rs, z, tp(rs, z, xt))
#
#
# def f(rs, z, xt, xs0, xs1):
#     return f_pbe(rs, z, xt, xs0, xs1)


# func = f(rs, z, xt, xs0, xs1)
# print(func)
# print(sp.ccode(func))


# https://gitlab.com/libxc/libxc/-/blob/master/maple/gga_exc/gga_x_pbe.mpl

# (* standard PBE *)
# $ifdef gga_x_pbe_params
params_a_kappa = sp.S(0.8040)
# params_a_mu = sp.S(0.2195149727645171)
params_a_mu = sp.S(0.249)  # TODO: why?


# # (* PBE_SOL *)
# # $ifdef gga_x_pbe_sol_params
# params_a_kappa = sp.S(0.8040)
# params_a_mu = MU_GE
#
# # $ifdef gga_x_pbe_tca_params
# params_a_kappa = sp.S(1.227)
# params_a_mu = sp.S(0.2195149727645171)


def pbe_f0(s):
    return 1 + params_a_kappa * (1 - params_a_kappa / (params_a_kappa + params_a_mu * s ** 2))


def pbe_f(x):
    return pbe_f0(X2S * x)


def f(rs, z, xt, xs0, xs1):
    return gga_exchange(pbe_f, rs, z, xs0, xs1)


# https://gitlab.com/libxc/libxc/-/blob/master/scripts/maple2c.pl

# zk is energy per unit particle
def mzk(r0, r1, s0, s1, s2):
    return f(r_ws(dens(r0, r1)), zeta(r0, r1), xt(r0, r1, s0, s1, s2), xs0(r0, r1, s0, s2), xs1(r0, r1, s0, s2))


# mf is energy per unit volume
def mf(r0, r1, s0, s1, s2):
    return dens(r0, r1) * mzk(r0, r1, s0, s1, s2)


unpol = False
if unpol:
    def dens(r0, r1):
        return r0


    def zeta(r0, r1):
        return 0


    def xs0(r0, r1, sigma0, sigma2):
        return sqrt(sigma0 / 4) / ((r0 / 2) ** (1 + sp.S(1 / DIMENSIONS)))


    def xs1(r0, r1, sigma0, sigma2):
        return sqrt(sigma0 / 4) / ((r0 / 2) ** (1 + sp.S(1 / DIMENSIONS)))


    def xt(r0, r1, sigma0, sigma1, sigma2):
        return sqrt(sigma0) / r0 ** (1 + sp.S(1 / DIMENSIONS))


else:  # pol
    def dens(r0, r1):
        return r0 + r1


    def zeta(r0, r1):
        return (r0 - r1) / (r0 + r1)


    def xs0(r0, r1, sigma0, sigma2):
        return sqrt(sigma0) / r0 ** (1 + sp.S(1 / DIMENSIONS))


    def xs1(r0, r1, sigma0, sigma2):
        return sqrt(sigma2) / r1 ** (1 + sp.S(1 / DIMENSIONS))


    def xt(r0, r1, sigma0, sigma1, sigma2):
        return sqrt(sigma0 + 2 * sigma1 + sigma2) / (r0 + r1) ** (1 + sp.S(1 / DIMENSIONS))

# func = f(rs, z, xt, xs0, xs1)
r0, r1, s0, s1, s2 = sp.symbols("r0, r1, s0, s1, s2")

func_mzk = mzk(r0, r1, s0, s1, s2)
func_mf = mf(r0, r1, s0, s1, s2)
func_mf_der_r0 = func_mf.diff(r0)
func_mf_der_r1 = func_mf.diff(r1)
func_mf_der_s0 = func_mf.diff(s0)
func_mf_der_s1 = func_mf.diff(s1)
func_mf_der_s2 = func_mf.diff(s2)

# func_mzk = sp.simplify(func_mzk)
# func_mf = sp.simplify(func_mf)
# func_mf_der_r0 = sp.simplify(func_mf_der_r0)
# func_mf_der_r1 = sp.simplify(func_mf_der_r1)
# func_mf_der_s0 = sp.simplify(func_mf_der_s0)
# func_mf_der_s1 = sp.simplify(func_mf_der_s1)
# func_mf_der_s2 = sp.simplify(func_mf_der_s2)

unknowns = ["izk[0]", "ivrho[0]", "ivrho[1]", "ivsigma[0]", "ivsigma[1]", "ivsigma[2]"]
updates = [func_mzk, func_mf, func_mf_der_r0, func_mf_der_r1, func_mf_der_s0, func_mf_der_s1, func_mf_der_s2]

print(f"""
double mzk (double r0, double r1, double s0, double s1, double s2, double p_a_zeta_threshold, double p_a_dens_threshold) {'{'}
    return {sp.ccode(func_mzk)};
{'}'}

double func_mf_der_r0 (double r0, double r1, double s0, double s1, double s2, double p_a_zeta_threshold, double p_a_dens_threshold) {'{'}
    return {sp.ccode(func_mf_der_r0)};
{'}'}

double func_mf_der_r1 (double r0, double r1, double s0, double s1, double s2, double p_a_zeta_threshold, double p_a_dens_threshold) {'{'}
    return {sp.ccode(func_mf_der_r1)};
{'}'}

double func_mf_der_s0 (double r0, double r1, double s0, double s1, double s2, double p_a_zeta_threshold, double p_a_dens_threshold) {'{'}
    return {sp.ccode(func_mf_der_s0)};
{'}'}

double func_mf_der_s1 (double r0, double r1, double s0, double s1, double s2, double p_a_zeta_threshold, double p_a_dens_threshold) {'{'}
    return {sp.ccode(func_mf_der_s1)};
{'}'}

double func_mf_der_s2 (double r0, double r1, double s0, double s1, double s2, double p_a_zeta_threshold, double p_a_dens_threshold) {'{'}
    return {sp.ccode(func_mf_der_s2)};
{'}'}
""")

# my_rho[0], my_rho[1], my_sigma[0], my_sigma[1], my_sigma[2], p->zeta_threshold, p->dens_threshold
var_mapping = {r0: sp.Symbol("my_rho[0]"), r1: sp.Symbol("my_rho[1]"),
               s0: sp.Symbol("my_sigma[0]"), s1: sp.Symbol("my_sigma[1]"), s2: sp.Symbol("my_sigma[2]"),
               p_a_zeta_threshold: sp.Symbol("p->zeta_threshold"), p_a_dens_threshold: sp.Symbol("p->dens_threshold")}

# print(f"""
# void func_pol_gen(
#     const xc_func_type *p, int order, const double *my_rho, const double *my_sigma, double *izk,
#     double *ivrho, double *ivsigma, double *iv2rho2, double *iv2rhosigma, double *iv2sigma2,
#     double *iv3rho3, double *iv3rho2sigma, double *iv3rhosigma2, double *iv3sigma3,
#     double *iv4rho4, double *iv4rho3sigma, double *iv4rho2sigma2, double *iv4rhosigma3, double *iv4sigma4) {'{'}
#     izk[0] = {sp.ccode(func_mzk.subs(var_mapping))};
#     ivrho[0] = {sp.ccode(func_mf_der_r0.subs(var_mapping))};
#     ivrho[1] = {sp.ccode(func_mf_der_r1.subs(var_mapping))};
#     ivsigma[0] = {sp.ccode(func_mf_der_s0.subs(var_mapping))};
#     ivsigma[1] = {sp.ccode(func_mf_der_s1.subs(var_mapping))};
#     ivsigma[2] = {sp.ccode(func_mf_der_s2.subs(var_mapping))};
# {'}'}
# """)

replacements, reduced_updates = sp.cse(updates)

print(f"""
void func_pol_gen(
    const xc_func_type *p, int order, const double *my_rho, const double *my_sigma, double *izk,
    double *ivrho, double *ivsigma, double *iv2rho2, double *iv2rhosigma, double *iv2sigma2,
    double *iv3rho3, double *iv3rho2sigma, double *iv3rhosigma2, double *iv3sigma3,
    double *iv4rho4, double *iv4rho3sigma, double *iv4rho2sigma2, double *iv4rhosigma3, double *iv4sigma4) {'{'}""")

for cs in replacements:
    print(f"    const double {sp.ccode(cs[0])} = {sp.ccode(cs[1].subs(var_mapping))};")

for unknown, update in zip(unknowns, reduced_updates):
    print(f"    {unknown} = {sp.ccode(update.subs(var_mapping))};")

print(f"""{'}'}
""")
